/*************************************************
 * File: test-harness.cpp 
 * Author: Keith Schwarz (htiek@cs.stanford.edu)
 *
 * File containing several test cases that can be
 * used to verify the correctness of the KDTree
 * implementation.  While I've tried to exercise
 * as many aspects of the KDTree as possible here,
 * there are almost certainly cases I haven't
 * thought of.  You should make sure to do your own
 * testing in addition to ensuring that the test
 * cases here pass.
 */
#include <iostream>
#include <string>
#include <sstream>
#include <vector>
#include <iomanip>
#include <cstdarg>
#include <set>
#include "../KDTree.h"
using namespace std;

/* These flags control which tests will be run.  Initially, only the
 * basic test will be executed.  As you complete more and more parts
 * of the implementation, you will want to turn more and more of these
 * flags on.
 */
#define BasicKDTreeTestEnabled            1 // Step two checks
#define ModerateKDTreeTestEnabled         1
#define HarderKDTreeTestEnabled           1
#define EdgeCaseKDTreeTestEnabled         1
    
#define NearestNeighborTestEnabled        1 // Step three checks
#define MoreNearestNeighborTestEnabled    1

#define BasicCopyTestEnabled              0 // Step four checks
#define ModerateCopyTestEnabled           0

/* Function which constructs a pair of a string and a point in space.  This uses
 * C++'s variable-length argument feature.  I _strongly_ encourage you never to
 * write code like this, as it's extremely error-prone, but for the purposes of
 * this test suite it ended up working out quite nicely.
 */
KDTree::value_type MakeValueType(const string& label, size_t numArgs, ...)
{
    /* Store the label. */
    KDTree::value_type result;
    result.second = label;

    /* Traverse the list of variable arguments. */
    va_list args;
    va_start(args, numArgs);

    for (size_t i = 0; i < numArgs; ++i)
        result.first.push_back(va_arg(args, double));

    va_end(args);

    return result;
}

/* Similar to above, except that it only generates a vector<double> */
vector<double> Vec(size_t numArgs, ...)
{
    vector<double> result;

    va_list args;
    va_start(args, numArgs);

    for (size_t i = 0; i < numArgs; ++i)
        result.push_back(va_arg(args, double));

    va_end(args);

    return result;
}

/* Utility function to convert an integer to a string. */
string IntegerToString(size_t i)
{
    stringstream converter;
    converter << i;
    return converter.str();
}

/* This function is what the test suite uses to ensure that the KDTree works
 * correctly.  It takes as parameters an expression and description, along
 * with a file and line number, then checks whether the condition is true.
 * If so, it prints that the test passed.  Otherwise, it reports that the
 * test fails and points the caller to the proper file and line.
 */
void DoCheckCondition(bool expr, const string& rationale, const string& file, int line)
{
    /* It worked!  Congrats. */
    if (expr)
    {
        cout << "PASS: " << rationale << endl;
        return;
    }

    /* Uh oh!  The test failed! */
    cout << "FAIL: " << rationale << endl;
    cout << "  Error at " << file << ", line " << line << endl;
    cout << "  (ENTER to continue)" << endl;

    /* Pause so that the test fail stands out. */
    string unused;
    getline(cin, unused);
}

/* This macro takes in an expression and a string, then invokes
 * DoCheckCondition passing in the arguments along with the file
 * and line number on which the macro was called.  This makes it
 * easier to track down the source of bugs if a test case should
 * fail.
 */
#define CheckCondition(expr, rationale) DoCheckCondition(expr, rationale, __FILE__, __LINE__)

/* Utility function to delimit the start and end of test cases. */
void PrintBanner(const string& header)
{
    cout << "\nBeginning test: " << header << endl;
    cout << setw(40) << setfill('-') << "" << setfill(' ') << endl;
}

/* Utility function to signal that a test isn't begin run. */
void TestDisabled(const string& header)
{
    cout << "== Test " << header << " NOT RUN: press ENTER to continue ==" << endl;

    string unused;
    getline(cin, unused);
}

/* Utility function to signal the end of a test. */
void EndTest()
{
    cout << "== end of test: press ENTER to continue ==" << endl;

    string unused;
    getline(cin, unused);
}

/* Basic test: Can we build a small tree and look up the elements it contains? */
void BasicKDTreeTest()
{
#if BasicKDTreeTestEnabled
    PrintBanner("Basic KDTree Test");

    /* Build the initial data set. */
    vector<KDTree::value_type> values;
    double dataPoints[3][3] = { 
        {1, 0, 0},
        {0, 1, 0},
        {0, 0, 1}};
    for (size_t i = 0; i < 3; ++i)
        values.push_back(KDTree::value_type(vector<double>(dataPoints[i], dataPoints[i] + 3),
                                            IntegerToString(i)));

    /* Construct the KDTree. */
    KDTree kd(values.begin(), values.end());
    CheckCondition(true, "KDTree construction completed.");

    /* Check basic properties of the KDTree. */
    CheckCondition(kd.dimension() == 3, "Dimension is three.");
    CheckCondition(kd.size() == 3, "New KD tree has the right number of elements.");
    CheckCondition(!kd.empty(),    "New KD tree is nonempty.");

    /* Make sure that the elements we built the tree out of are still there. */
    CheckCondition(kd.find(values[0].first) != kd.end(), "New KD tree has element zero.");
    CheckCondition(kd.find(values[1].first) != kd.end(), "New KD tree has element one.");
    CheckCondition(kd.find(values[2].first) != kd.end(), "New KD tree has element two.");

    /* Note: The !! syntax means "convert to bool".  It works since !x is true if x is nonzero
     * and false otherwise, and !!x then inverts the result.
     */
    CheckCondition(!!kd.count(values[0].first), "count agrees with find.");
    CheckCondition(!!kd.count(values[1].first), "count agrees with find.");
    CheckCondition(!!kd.count(values[2].first), "count agrees with find.");

    /* Check that the key/value pairs were correctly preserved. */
    CheckCondition(*kd.find(values[0].first) == values[0], "Key/value pairs preserved.");

    /* Now, iterate over the KDTree and ensure that:
     * 1. The right number of elements are there
     * 2. Nothing else is.
     */
    set<KDTree::value_type> reference(values.begin(), values.end());
    set<KDTree::value_type> actual(kd.begin(), kd.end());
    CheckCondition(reference.size() == actual.size(), "Iteration visits right number of elements.");
    CheckCondition(equal(reference.begin(), reference.end(), actual.begin()), "Iteration visits correct elements.");

    /* Try looking up some nonexistent elements and see what happens. */
    CheckCondition(!kd.count(Vec(3, 1.0, 1.0, 1.0)), "Nonexistent elements aren't in the tree.");
    CheckCondition(!kd.count(Vec(3, 0.0, 0.0, 0.0)), "Nonexistent elements aren't in the tree.");

    EndTest();
#else
    TestDisabled("BasicKDTreeTest");
#endif
}

/* Test is similar to BasicKDTreeTest, but the data set is bigger and a bit more complicated. */
void ModerateKDTreeTest()
{
#if ModerateKDTreeTestEnabled
    PrintBanner("Moderate KDTree Test");

    /* Build the data set. */
    vector<KDTree::value_type> values;
    double dataPoints[][4] = { 
        {0, 0, 0, 0},
        {0, 0, 0, 1},
        {0, 0, 1, 0},
        {0, 0, 1, 1},
        {0, 1, 0, 0},
        {0, 1, 0, 1},
        {0, 1, 1, 0},
        {0, 1, 1, 1},
        {1, 0, 0, 0},
        {1, 0, 0, 1},
        {1, 0, 1, 0},
        {1, 0, 1, 1},
        {1, 1, 0, 0},
        {1, 1, 0, 1},
        {1, 1, 1, 0},
        {1, 1, 1, 1},
    };
    for (size_t i = 0; i < 16; ++i)
        values.push_back(KDTree::value_type(vector<double>(dataPoints[i], dataPoints[i] + 4),
                                            IntegerToString(i)));

    /* Build the tree. */
    KDTree kd(values.begin(), values.end());

    /* Check that basic properties hold. */
    CheckCondition(kd.dimension() == 4, "Dimension is four.");
    CheckCondition(kd.size() == 16, "New KD tree has the right number of elements.");
    CheckCondition(!kd.empty(),    "New KD tree is nonempty.");

    /* Check that all of the elements we built the tree from are still present. */
    for (size_t i = 0; i < 16; ++i)
    {
        CheckCondition(kd.find(values[i].first) != kd.end(), "New KD tree has original elems.");
        CheckCondition(!!kd.count(values[i].first), "count agrees with find.");
    }

    /* Check that iteration still works. */
    set<KDTree::value_type> reference(values.begin(), values.end());
    set<KDTree::value_type> actual(kd.begin(), kd.end());
    CheckCondition(reference.size() == actual.size(), "Iteration visits right number of elements.");
    CheckCondition(equal(reference.begin(), reference.end(), actual.begin()), "Iteration visits correct elements.");

    /* Try looking up some nonexistent elements and see what happens. */
    CheckCondition(!kd.count(Vec(4, 1.0, 1.0, 1.0,  0.5)), "Nonexistent elements aren't in the tree.");
    CheckCondition(!kd.count(Vec(4, 0.0, 0.0, 0.0, -0.5)), "Nonexistent elements aren't in the tree.");

    EndTest();
#else
    TestDisabled("ModerateKDTreeTest");
#endif
}

/* This test still uses just the basic functionality, but it hammers it a bit more
 * by checking for strange edge cases like duplicated elements.
 */
void HarderKDTreeTest()
{
#if HarderKDTreeTestEnabled
    PrintBanner("Harder KDTree Test");

    /* Build the data set. */
    vector<KDTree::value_type> values;
    double dataPoints[][4] = { 
        {0, 0, 0, 0},
        {0, 1, 0, 1},
        {0, 0, 0, 0}, // Duplicate!
        {0, 1, 0, 1}, // Duplicate!
        {0, 1, 1, 0},
        {1, 0, 1, 0},        
    };
    for (size_t i = 0; i < 6; ++i)
        values.push_back(KDTree::value_type(vector<double>(dataPoints[i], dataPoints[i] + 4),
                                            IntegerToString(i)));

    /* Build the KDTree. */
    KDTree kd(values.begin(), values.end());

    /* Check basic properties. */
    CheckCondition(kd.dimension() == 4, "Dimension is four.");
    CheckCondition(kd.size() == 4, "New KD tree has the right number of elements.");
    CheckCondition(!kd.empty(),    "New KD tree is nonempty.");

    /* Make sure that elements are still there. */
    for (size_t i = 0; i < 6; ++i)
    {
        CheckCondition(kd.find(values[i].first) != kd.end(), "New KD tree has original elems.");
        CheckCondition(!!kd.count(values[i].first), "count agrees with find.");
    }

    /* Remove the duplicates from the test set and ensure that everything is still there. */
    values.erase(values.begin() + 3);
    values.erase(values.begin() + 2);
    set<KDTree::value_type> reference(values.begin(), values.end());
    set<KDTree::value_type> actual(kd.begin(), kd.end());
    CheckCondition(reference.size() == actual.size(), "Iteration visits right number of elements.");
    CheckCondition(equal(reference.begin(), reference.end(), actual.begin()), "Iteration visits correct elements.");

    EndTest();
#else
    TestDisabled("HarderKDTreeTest");
#endif
}

/* This test builds a KDTree where the data has the same values everywhere except 
 * along one coordinate axis.  If you are failing this test case, make sure that
 * your implementation of find() descends into the left subtree only if the current
 * coordinate is _strictly less_ than the partition point's coordinate.
 */
void EdgeCaseKDTreeTest()
{
#if EdgeCaseKDTreeTestEnabled
    PrintBanner("Edge Case KDTree Test");

    /* Build the data set. */
    vector<KDTree::value_type> values;
    double dataPoints[8][3] = { 
        {0, 0, 0},
        {0, 1, 0},
        {0, 2, 0},
        {0, 3, 0},
        {0, 4, 0},
        {0, 5, 0},
        {0, 6, 0},
        {0, 7, 0},
    };
    for (size_t i = 0; i < 8; ++i)
        values.push_back(KDTree::value_type(vector<double>(dataPoints[i], dataPoints[i] + 3),
                                            IntegerToString(i)));

    /* Convert it to a KDTree. */
    KDTree kd(values.begin(), values.end());

    /* Basic checks. */
    CheckCondition(kd.dimension() == 3, "Dimension is three.");
    CheckCondition(kd.size() == 8, "New KD tree has the right number of elements.");
    CheckCondition(!kd.empty(),    "New KD tree is nonempty.");

    /* Make sure everything can be found. */
    for (size_t i = 0; i < values.size(); ++i)
        CheckCondition(!!kd.count(values[i].first), "Lookup succeeded.");
    
    /* Iterate and check that nothing was lost. */
    set<KDTree::value_type> reference(values.begin(), values.end());
    set<KDTree::value_type> actual(kd.begin(), kd.end());
    CheckCondition(reference.size() == actual.size(), "Iteration visits right number of elements.");
    CheckCondition(equal(reference.begin(), reference.end(), actual.begin()), "Iteration visits correct elements.");

    EndTest();
#else
    TestDisabled("EdgeCaseTreeTest");
#endif
}

/* This is a simple test case that checks whether 1-NN lookups work
 * correctly.  If you're failing this test case, you probably need
 * to check that your implementation of kNearestNeighborValue works
 * correctly.
 */
void NearestNeighborTest()
{
#if NearestNeighborTestEnabled
    PrintBanner("Nearest Neighbor Test");

    /* Build two data sets - a set of tree points and a set of test points. */
    vector<KDTree::value_type> values;
    vector<KDTree::value_type> tests;
    double dataPoints[][4] = { 
        {0, 0, 0, 0}, // 0
        {0, 0, 0, 1}, // 1
        {0, 0, 1, 0}, // 2
        {0, 0, 1, 1}, // 3
        {0, 1, 0, 0}, // 4
        {0, 1, 0, 1}, // 5
        {0, 1, 1, 0}, // 6
        {0, 1, 1, 1}, // 7
        {1, 0, 0, 0}, // 8
        {1, 0, 0, 1}, // 9
        {1, 0, 1, 0}, // 10
        {1, 0, 1, 1}, // 11
        {1, 1, 0, 0}, // 12
        {1, 1, 0, 1}, // 13
        {1, 1, 1, 0}, // 14
        {1, 1, 1, 1}, // 15
    };

    /* Each test point should be right next to the corresponding point in the
     * tree point set.
     */
    double testPoints[][4] = { 
        {0, 0, 0, 0}, // 0
        {0, 0, 0, 0.7}, // 1
        {0, 0, 0.9, 0}, // 2
        {0, 0, 0.6, 0.6}, // 3
        {0, 0.9, 0, 0}, // 4
        {0, 0.8, 0, 0.7}, // 5
        {0, 0.6, 0.7, -0.1}, // 6
        {-0.4, 0.7, 0.7, 0.7}, // 7
        {1, 0, 0, 0}, // 8
        {1, 0, 0, 1}, // 9
        {1, 0, 1, 0}, // 10
        {1, 0, 1, 1}, // 11
        {1, 1, 0, 0}, // 12
        {1, 1, 0, 1}, // 13
        {1, 1, 1, 0}, // 14
        {1, 1, 1, 1}, // 15
    };

    /* Build the two sets. */
    for (size_t i = 0; i < 16; ++i)
        values.push_back(KDTree::value_type(vector<double>(dataPoints[i], dataPoints[i] + 4),
                                            IntegerToString(i)));
    for (size_t i = 0; i < 16; ++i)
        tests.push_back(KDTree::value_type(vector<double>(testPoints[i], testPoints[i] + 4),
                                            IntegerToString(i)));

    /* Construct the tree from the tree points. */
    KDTree kd(values.begin(), values.end());

    /* Check that calling nearest neighbor on a point in the tree yields that point. */
    for (size_t i = 0; i < 16; ++i)
        CheckCondition(kd.kNearestNeighborValue(values[i].first, 1) == values[i].second, "Nearest neighbor of element is that element.");

    /* Check that calling nearest neighbor on the test points yields the correct tree points. */
    for (size_t i = 0; i < 16; ++i)
        CheckCondition(kd.kNearestNeighborValue(tests[i].first, 1) == tests[i].second, "Tests pass.");

    EndTest();
#else
    TestDisabled("NearestNeighborTest");
#endif
}

/* This test verifies that k-NN works for k > 1.  If you're failing this test case, then you probably
 * need to check that you're computing the majority element correctly.
 */
void MoreNearestNeighborTest()
{
#if MoreNearestNeighborTestEnabled
    PrintBanner("More Nearest Neighbor Tests");

    /* We're going to build a 3x3 lattice that looks like this:
     *
     * a b a
     * b a b
     * b b b
     *
     * This will give us some fun checking whether k-NN works
     * correctly.
     */
    vector<KDTree::value_type> values;
    values.push_back(MakeValueType("a", 2, 0.0, 0.0));
    values.push_back(MakeValueType("b", 2, 0.0, 0.5));
    values.push_back(MakeValueType("a", 2, 0.0, 1.0));
    values.push_back(MakeValueType("b", 2, 0.5, 0.0));
    values.push_back(MakeValueType("a", 2, 0.5, 0.5));
    values.push_back(MakeValueType("b", 2, 0.5, 1.0));
    values.push_back(MakeValueType("b", 2, 1.0, 0.0));
    values.push_back(MakeValueType("b", 2, 1.0, 0.5));
    values.push_back(MakeValueType("b", 2, 1.0, 1.0));

    /* Build the tree. */
    KDTree kd(values.begin(), values.end());
    
    /* Look for the the four points closes to the top-middle.  This should give us 'a' even though
     * the point itself is 'b'.
     */
    CheckCondition(kd.kNearestNeighborValue(Vec(2, 0.0,   0.5),  4) == "a", "Nearest neighbors are correct.");

    /* Check for the nine points closest to the center.  This should give us 'b' even though the
     * point itself is 'a'
     */
    CheckCondition(kd.kNearestNeighborValue(Vec(2, 0.0,   0.5),  9) == "b", "Nearest neighbors are correct.");

    /* Check for the five points closest to the center.  This should give us 'b' even though the
     * point itself is 'a'
     */
    CheckCondition(kd.kNearestNeighborValue(Vec(2, 0.5,   0.5),  5) == "b", "Nearest neighbors are correct.");

    /* Look in the center of the square in the bottom-right corner.  All the points are 'b' and
     * so we'd better get 'b' back!
     */
    CheckCondition(kd.kNearestNeighborValue(Vec(2, 0.75, 0.75),  4) == "b", "Nearest neighbors are correct.");

    /* Choose a point way out of the box and get the 9-NN value.  Since there are only 9 data points,
     * this should be equivalent to asking "what's the most common element?"  (The answer is 'b')
     */
    CheckCondition(kd.kNearestNeighborValue(Vec(2, 10.0, 10.0),  9) == "b", "Nearest neighbor at distant point is majority element.");

    /* Similar check but for negative numbers. */
    CheckCondition(kd.kNearestNeighborValue(Vec(2, -5.0, -5.0),  9) == "b", "No problems with negative values.");

    /* Try getting the 25 nearest neighbors to a point.  There are only nine points, but this should
     * still work correctly.
     */
    CheckCondition(kd.kNearestNeighborValue(Vec(2, -5.0, -5.0), 25) == "b", "No problems with looking up more neighbors than elements.");

    EndTest();
#else
    TestDisabled("MoreNearestNeighborTest");
#endif
}

void BasicCopyTest()
{
#if BasicCopyTestEnabled
    PrintBanner("Basic Copy Test");

    /* Create two KDTrees as references.  One will have dimension three,
     * one will have dimension four.
     */
    vector<KDTree::value_type> points2d, points3d;
    points2d.push_back(MakeValueType("a", 2, 0.0, 0.0));
    points2d.push_back(MakeValueType("a", 2, 0.0, 1.0));
    points2d.push_back(MakeValueType("a", 2, 1.0, 0.0));
    points2d.push_back(MakeValueType("a", 2, 1.0, 1.0));

    points3d.push_back(MakeValueType("a", 3, 0.0, 0.0, 0.0));
    points3d.push_back(MakeValueType("a", 3, 0.0, 0.0, 5.0));
    points3d.push_back(MakeValueType("a", 3, 0.0, 5.0, 0.0));
    points3d.push_back(MakeValueType("a", 3, 0.0, 5.0, 5.0));
    points3d.push_back(MakeValueType("a", 3, 5.0, 0.0, 5.0));
    points3d.push_back(MakeValueType("a", 3, 5.0, 0.0, 5.0));
    points3d.push_back(MakeValueType("a", 3, 5.0, 5.0, 5.0));
    points3d.push_back(MakeValueType("a", 3, 5.0, 5.0, 5.0));

    /* Build them into kd-trees. */
    const KDTree one(points2d.begin(), points2d.end());
    const KDTree two(points3d.begin(), points3d.end());

    {
        /* Test one: Make a copy of one and ensure it works correctly. */
        KDTree clone = one;
        CheckCondition(clone.dimension() == 2, "Dimension is two.");
        CheckCondition(clone.size() == one.size(), "Copied size should agree.");
        CheckCondition(clone.empty() == one.empty(), "Copied empty should agree.");
        
        for (size_t i = 0; i < points2d.size(); ++i)
            CheckCondition(clone.find(points2d[i].first) != clone.end(), "Copy should hold everything from the master set.");
    }

    {
        /* Test two: Make a copy of two and ensure it works correctly. */
        KDTree clone = two;
        CheckCondition(clone.dimension() == 3, "Dimension is three.");
        CheckCondition(clone.size() == two.size(), "Copied size should agree.");
        CheckCondition(clone.empty() == two.empty(), "Copied empty should agree.");
        
        for (size_t i = 0; i < points3d.size(); ++i)
            CheckCondition(clone.find(points3d[i].first) != clone.end(), "Copy should hold everything from the master set.");
    }

    {
        /* Test three: Make a KDTree that's a copy of one and copy two on top of it. */
        KDTree clone = one;
        clone = two;

        CheckCondition(clone.dimension() == two.dimension(), "Dimension is three.");
        CheckCondition(clone.size() == two.size(), "Copied size should agree.");
        CheckCondition(clone.empty() == two.empty(), "Copied empty should agree.");
        
        for (size_t i = 0; i < points3d.size(); ++i)
            CheckCondition(clone.find(points3d[i].first) != clone.end(), "Copy should hold everything from the master set.");
    }
    EndTest();
#else
    TestDisabled("BasicCopyTest");
#endif
}

void ModerateCopyTest()
{
#if ModerateCopyTestEnabled
    PrintBanner("Moderate Copy Test");

    /* Create two KDTrees as references.  Both have dimension two. */
    vector<KDTree::value_type> pointsOne, pointsTwo;
    pointsOne.push_back(MakeValueType("a", 2, 0.0, 0.0));
    pointsOne.push_back(MakeValueType("a", 2, 0.0, 1.0));
    pointsOne.push_back(MakeValueType("a", 2, 1.0, 0.0));
    pointsOne.push_back(MakeValueType("a", 2, 1.0, 1.0));

    pointsTwo.push_back(MakeValueType("a", 2, 0.0, 2.0));
    pointsTwo.push_back(MakeValueType("a", 2, 0.0, 3.0));
    pointsTwo.push_back(MakeValueType("a", 2, 0.0, 4.0));
    pointsTwo.push_back(MakeValueType("a", 2, 0.0, 5.0));
    pointsTwo.push_back(MakeValueType("a", 2, 0.0, 6.0));
    pointsTwo.push_back(MakeValueType("a", 2, 0.0, 7.0));
    pointsTwo.push_back(MakeValueType("a", 2, 0.0, 8.0));
    pointsTwo.push_back(MakeValueType("a", 2, 0.0, 9.0));
    
    /* Note that pointsOne and pointsTwo are disjoint. */

    const KDTree one(pointsOne.begin(), pointsOne.end());
    const KDTree two(pointsTwo.begin(), pointsTwo.end());

    /* Exercise common edge cases of the assignment operator.  If you get a compile-time
     * error when executing this test, make sure that your implementation of the assignment
     * operator correctly returns a KDTree&.
     */
    KDTree clone = one;
    clone = clone = clone = clone; // Chained self-assignment.
    (clone = two) = clone;         // Assignment should return lvalue.
    (clone = clone) = (clone = one) = (clone = two); // Chained nested assignments
    clone = two;
    clone = clone; // Basic self-assignment.

    /* At this point clone should be a copy of two.  We need to ensure that:
     * 1. clone behaves like two.
     * 2. one didn't change.
     * 3. two didn't change.
     */
    CheckCondition(one.size() == pointsOne.size(), "one still has the right number of elements.");
    CheckCondition(two.size() == pointsTwo.size(), "two still has the right number of elements.");
    CheckCondition(clone.size() == two.size(), "clone has the same size as two.");
    CheckCondition(one.dimension() == 2, "Dimension is unchanged.");
    CheckCondition(two.dimension() == 2, "Dimension is unchanged.");
    CheckCondition(clone.dimension() == 2, "clone has right dimension.");
    
    /* Check that one and two still hold. */
    for (size_t i = 0; i < pointsOne.size(); ++i)
    {
        CheckCondition(!!one.count(pointsOne[i].first), "one holds the right value.");
        CheckCondition(!two.count(pointsOne[i].first), "two doesn't hold points from one.");
        CheckCondition(!clone.count(pointsOne[i].first), "clone doesn't hold points from one.");
    }
    for (size_t i = 0; i < pointsTwo.size(); ++i)
    {
        CheckCondition(!one.count(pointsTwo[i].first), "one doesn't hold points from two.");
        CheckCondition(!!two.count(pointsTwo[i].first), "two holds the correct value.");
        CheckCondition(!!clone.count(pointsTwo[i].first), "clone holds the correct value.");
    }

    /* Check that the containers are actually equal. */
    CheckCondition(equal(two.begin(), two.end(), clone.begin()), "two == clone");
    EndTest();
#else
    TestDisabled("ModerateCopyTest");
#endif
}

/* Main entry point simply runs all the tests.  Note that these functions might be no-ops
 * if they are disabled by the configuration settings at the top of the program.
 */
int main()
{
    /* Step Two Tests */
    BasicKDTreeTest();
    ModerateKDTreeTest();
    HarderKDTreeTest();
    EdgeCaseKDTreeTest();
    
    /* Step Three Tests */
    NearestNeighborTest();
    MoreNearestNeighborTest();

    /* Step Four Tests */
    BasicCopyTest();
    ModerateCopyTest();

    #if (BasicKDTreeTestEnabled && \
         ModerateKDTreeTestEnabled && \
         HarderKDTreeTestEnabled && \
         EdgeCaseKDTreeTestEnabled && \
         NearestNeighborTestEnabled && \
         MoreNearestNeighborTestEnabled && \
         BasicCopyTestEnabled && \
         ModerateCopyTestEnabled)
        cout << "All tests completed!  If they passed, you should be good to go!" << endl << endl;
    #else
        cout << "Not all tests were run.  Enable the rest of the tests, then run again." << endl << endl;
    #endif

    EndTest();
}
